from dataclasses import dataclass


@dataclass
class InfoData:
    # /cgi/info.js data
    device_description: str | None
    display_serial: str | None
    sw_id_ems: int | None
    hw_id_ems: int | None
    countrycode: int | None
    sw_version_ems: str | None
    anz_charger: int | None
    soll_charger: int | None
    serial_emeter: str | None
    mac_emeter: str | None
    sw_version_emeter: str | None
    bl_version_emeter: str | None
    hw_id_emeter: int | None
    serial_wr: str | None
    mac_wr: int | None
    sw_id_wr: int | None
    hw_id_wr: int | None
    sw_version_wr: str | None
    bl_version_wr: str | None
    serial_ens: str | None
    mac_ens: int | None
    sw_id_ens: int | None
    hw_id_ens: int | None
    sw_version_ens: str | None
    bl_version_ens: str | None
    charger_serial: list[str]
    charger_mac: list[str]
    sw_id_charger: list[int]
    hw_id_charger: list[int]
    sw_version_charger: list[str]
    bl_version_charger: list[str]
    p_ems_max: int | None
    p_ems_maxdisc: int | None
    battery_sw: list[str]
    battery_hw: list[str]
    battery_serial: list[str]
    bm_update: list[str]
    bm_update_sw: list[str]
    bm_production: list[str]
    lg_battery_serial: list[str]

    @classmethod
    def from_dict(cls, info: dict) -> "InfoData":
        return cls(
            device_description=info.get("Device_Description"),
            display_serial=info.get("Display_Serial"),
            sw_id_ems=info.get("SW_ID_EMS"),
            hw_id_ems=info.get("HW_ID_EMS"),
            countrycode=info.get("countrycode"),
            sw_version_ems=info.get("SW_Version_EMS"),
            anz_charger=info.get("Anz_Charger"),
            soll_charger=info.get("Soll_Charger"),
            serial_emeter=info.get("Serial_EMeter"),
            mac_emeter=info.get("MAC_EMeter"),
            sw_version_emeter=info.get("SW_Version_EMeter"),
            bl_version_emeter=info.get("BL_Version_EMeter"),
            hw_id_emeter=info.get("HW_ID_EMeter"),
            serial_wr=info.get("Serial_WR"),
            mac_wr=info.get("MAC_WR"),
            sw_id_wr=info.get("SW_ID_WR"),
            hw_id_wr=info.get("HW_ID_WR"),
            sw_version_wr=info.get("SW_Version_WR"),
            bl_version_wr=info.get("BL_Version_WR"),
            serial_ens=info.get("Serial_ENS"),
            mac_ens=info.get("MAC_ENS"),
            sw_id_ens=info.get("SW_ID_ENS"),
            hw_id_ens=info.get("HW_ID_ENS"),
            sw_version_ens=info.get("SW_Version_ENS"),
            bl_version_ens=info.get("BL_Version_ENS"),
            charger_serial=info.get("Charger_Serial", []),
            charger_mac=info.get("Charger_MAC", []),
            sw_id_charger=info.get("SW_ID_Charger", []),
            hw_id_charger=info.get("HW_ID_Charger", []),
            sw_version_charger=info.get("SW_Version_Charger", []),
            bl_version_charger=info.get("BL_Version_Charger", []),
            p_ems_max=info.get("P_EMS_Max"),
            p_ems_maxdisc=info.get("P_EMS_MaxDisc"),
            battery_sw=info.get("BatterySW", []),
            battery_hw=info.get("BatteryHW", []),
            battery_serial=info.get("BatterySerial", []),
            bm_update=info.get("BM_Update", []),
            bm_update_sw=info.get("BM_UpdateSW", []),
            bm_production=info.get("BM_Production", []),
            lg_battery_serial=info.get("LG_Battery_Serial", []),
        )


@dataclass
class EnergyData:
    # /cgi/energy.js data
    total_grid_ac_dc: int | None  # Wh
    total_grid_dc_ac: int | None  # Wh
    total_inverter_ac_dc: int | None  # Wh
    total_inverter_dc_ac: int | None  # Wh
    total_charge_cycles: list[int]

    @classmethod
    def from_dict(cls, energy: dict) -> "EnergyData":
        return cls(
            total_grid_ac_dc=energy.get("EGrid_AC_DC"),
            total_grid_dc_ac=energy.get("EGrid_DC_AC"),
            total_inverter_ac_dc=energy.get("EWr_AC_DC"),
            total_inverter_dc_ac=energy.get("EWr_DC_AC"),
            total_charge_cycles=energy.get("Chrg_LoadCycles", []),
        )


@dataclass
class ServiceData:
    # /cgi/user_serv.js data
    hours_until_filter_maintenance: int | None  # Hours
    status_fan: int | None
    status_main: int | None

    @classmethod
    def from_dict(cls, service: dict) -> "ServiceData":
        return cls(
            hours_until_filter_maintenance=service.get("FilterZeit"),
            status_fan=service.get("Fan"),
            status_main=service.get("Main"),
        )


@dataclass
class WrData:
    nominal_power: int | None  # W
    u_verbund_l1: int | None  # V
    u_verbund_l2: int | None  # V
    u_verbund_l3: int | None  # V
    i_verbund_l1: int | None  # A
    i_verbund_l2: int | None  # A
    i_verbund_l3: int | None  # A
    u_insel_l1: int | None  # V
    u_insel_l2: int | None  # V
    u_insel_l3: int | None  # V
    i_insel_l1: int | None  # A
    i_insel_l2: int | None  # A
    i_insel_l3: int | None  # A
    temp_l1: int | None  # Celcius
    temp_l2: int | None  # Celcius
    temp_l3: int | None  # Celcius
    temp_board: int | None  # Celcius
    frequency_grid: int | None  # Hz
    online_status: int | None  # 0=Offline, 1=Online
    fan_speed: int | None  # percentage

    @classmethod
    def from_dict(cls, wr: dict) -> "WrData":
        return cls(
            nominal_power=wr.get("PSoll"),
            u_verbund_l1=wr.get("U Verbund L1"),
            u_verbund_l2=wr.get("U Verbund L2"),
            u_verbund_l3=wr.get("U Verbund L3"),
            i_verbund_l1=wr.get("I Verbund L1"),
            i_verbund_l2=wr.get("I Verbund L2"),
            i_verbund_l3=wr.get("I Verbund L3"),
            u_insel_l1=wr.get("U Insel L1"),
            u_insel_l2=wr.get("U Insel L2"),
            u_insel_l3=wr.get("U Insel L3"),
            i_insel_l1=wr.get("I Insel L1"),
            i_insel_l2=wr.get("I Insel L2"),
            i_insel_l3=wr.get("I Insel L3"),
            temp_l1=wr.get("Temp L1"),
            temp_l2=wr.get("Temp L2"),
            temp_l3=wr.get("Temp L3"),
            temp_board=wr.get("TBoard"),
            frequency_grid=wr.get("FNetz"),
            online_status=wr.get("OnlineStatus"),
            fan_speed=wr.get("Luefter"),
        )


@dataclass
class EMeterData:
    f_netz: int | None
    sens_state: int | None
    u_v_l1: int | None
    u_v_l2: int | None
    u_v_l3: int | None
    iw_v_l1: int | None
    iw_v_l2: int | None
    iw_v_l3: int | None
    ib_v_l1: int | None
    ib_v_l2: int | None
    ib_v_l3: int | None
    is_v_l1: int | None
    is_v_l2: int | None
    is_v_l3: int | None
    iw_pv_l1: int | None
    iw_pv_l2: int | None
    iw_pv_l3: int | None
    ib_pv_l1: int | None
    ib_pv_l2: int | None
    ib_pv_l3: int | None
    is_pv_l1: int | None
    is_pv_l2: int | None
    is_pv_l3: int | None

    @classmethod
    def from_dict(cls, emeter: dict) -> "EMeterData":
        return cls(
            f_netz=emeter.get("FNetz"),
            sens_state=emeter.get("SensState"),
            u_v_l1=emeter.get("U_V_L1"),
            u_v_l2=emeter.get("U_V_L2"),
            u_v_l3=emeter.get("U_V_L3"),
            iw_v_l1=emeter.get("Iw_V_L1"),
            iw_v_l2=emeter.get("Iw_V_L2"),
            iw_v_l3=emeter.get("Iw_V_L3"),
            ib_v_l1=emeter.get("Ib_V_L1"),
            ib_v_l2=emeter.get("Ib_V_L2"),
            ib_v_l3=emeter.get("Ib_V_L3"),
            is_v_l1=emeter.get("Is_V_L1"),
            is_v_l2=emeter.get("Is_V_L2"),
            is_v_l3=emeter.get("Is_V_L3"),
            iw_pv_l1=emeter.get("Iw_PV_L1"),
            iw_pv_l2=emeter.get("Iw_PV_L2"),
            iw_pv_l3=emeter.get("Iw_PV_L3"),
            ib_pv_l1=emeter.get("Ib_PV_L1"),
            ib_pv_l2=emeter.get("Ib_PV_L2"),
            ib_pv_l3=emeter.get("Ib_PV_L3"),
            is_pv_l1=emeter.get("Is_PV_L1"),
            is_pv_l2=emeter.get("Is_PV_L2"),
            is_pv_l3=emeter.get("Is_PV_L3"),
        )


@dataclass
class EnsData:
    f_netz: int | None
    u_v_l1: int | None
    u_v_l2: int | None
    u_v_l3: int | None

    @classmethod
    def from_dict(cls, ens: dict) -> "EnsData":
        return cls(
            f_netz=ens.get("FNetz"),
            u_v_l1=ens.get("U_V_L1"),
            u_v_l2=ens.get("U_V_L2"),
            u_v_l3=ens.get("U_V_L3"),
        )


@dataclass
class ChargerData:
    # TODO
    pass


@dataclass
class BattData:
    # TODO
    pass



import time
from dataclasses import dataclass

from pymodbus.client.tcp import ModbusTcpClient
from pymodbus.constants import Endian
from pymodbus.exceptions import ModbusException
from pymodbus.payload import BinaryPayloadDecoder

ERROR_TEMPLATE = (
    "An error occured while polling address {}. "
    + "This might be an issue with your device."
)

CACHE_TIME = 900  # 15 minutes


@dataclass
class ModbusData:
    soc: int
    grid_power: int
    state: int
    active_power: int
    apparent_power: int
    error_code: int
    total_charged_energy: int
    number_modules: int
    installed_capacity: int
    serial: str
    table_version: int
    software_version_ems: str
    software_version_ens: str
    software_version_inverter: str


@dataclass
class CacheData:
    timestamp_cache: int = 0
    serial: str = ""
    table_version: int = 0
    software_version_ems: str = ""
    software_version_ens: str = ""
    software_version_inverter: str = ""

    def set_data(
        self,
        serial: str,
        table_version: int,
        software_version_ems: str,
        software_version_ens: str,
        software_version_inverter: str,
    ) -> None:
        self.timestamp_cache = int(time.time())
        self.serial = serial
        self.table_version = table_version
        self.software_version_ems = software_version_ems
        self.software_version_ens = software_version_ens
        self.software_version_inverter = software_version_inverter


class ModbusClient:
    def __init__(self, modbus_host, modbus_port) -> None:
        self._slave = 255

        self.modbus_host = modbus_host
        self.modbus_port = modbus_port
        self._modbus_client = ModbusTcpClient(
            host=self.modbus_host, port=self.modbus_port
        )

        self._cache = CacheData()

    def connect(self) -> bool:
        return self._modbus_client.connect()

    def disconnect(self) -> None:
        return self._modbus_client.close()

    def is_connected(self) -> bool:
        return self._modbus_client.is_socket_open()

    def get_all_data_modbus(self) -> ModbusData:
        self.update_cache()
        out = ModbusData(
            soc=self.get_soc(),
            grid_power=self.get_grid_power(),
            state=self.get_state(),
            active_power=self.get_active_power(),
            apparent_power=self.get_apparent_power(),
            error_code=self.get_error_code(),
            number_modules=self.get_bm_installed(),
            installed_capacity=self.get_installed_capacity(),
            total_charged_energy=self.get_total_charged_energy(),
            serial=self._cache.serial,
            table_version=self._cache.table_version,
            software_version_ems=self._cache.software_version_ems,
            software_version_ens=self._cache.software_version_ens,
            software_version_inverter=self._cache.software_version_inverter,
        )
        return out

    def update_cache(self) -> None:
        if int(time.time()) - self._cache.timestamp_cache < CACHE_TIME:
            # cache is still relevant
            return

        self._cache.set_data(
            serial=self.get_serial(),
            table_version=self.get_table_version(),
            software_version_ems=self.get_software_version_ems(),
            software_version_ens=self.get_software_version_ens(),
            software_version_inverter=self.get_software_version_inverter(),
        )

    def get_software_version_ems(self) -> str:
        registers = self._get_value_modbus(1000, 17)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.BIG
        ).decode_string(17)

        return self._clean_string(result)

    def get_software_version_ens(self) -> str:
        registers = self._get_value_modbus(1017, 17)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.BIG
        ).decode_string(17)

        return self._clean_string(result)

    def get_software_version_inverter(self) -> str:
        registers = self._get_value_modbus(1034, 17)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.BIG
        ).decode_string(17)

        return self._clean_string(result)

    def get_table_version(self) -> int:
        registers = self._get_value_modbus(1051, 1)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.LITTLE
        ).decode_16bit_uint()
        return result

    def get_serial(self) -> str:
        # Retrieves the Serial Number of the device
        # Supported on VARTA element, pulse, pulse neo, link and flex storage devices

        registers = self._get_value_modbus(1054, 10)

        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.BIG
        ).decode_string(18)

        return self._clean_string(result)

    def get_bm_installed(self) -> int:
        # Retrieves the number of battery modules installed
        # Supported on VARTA element, pulse, pulse neo, link and flex storage devices

        registers = self._get_value_modbus(1064, 1)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.LITTLE
        ).decode_16bit_uint()
        return result

    def get_state(self) -> int:
        # Retrieves the state of the device
        #  # "BUSY" (e.g. during startup) = 0/ "RUN" (ready to charge / discharge) = 1/
        # "CHARGE" = 2/ "DISCHARGE" = 3/ "STANDBY" = 4 /"ERROR" = 5 /
        # "PASSIVE" (service) = 6/ "ISLANDING" = 7
        # Supported on VARTA element, pulse, pulse neo, link and flex storage devices

        registers = self._get_value_modbus(1065, 1)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.LITTLE
        ).decode_16bit_uint()
        return result

    def get_active_power(self) -> int:
        # Active Power measured at the internal inverter. Positive = Charge,
        # Negative = Discharge
        # Supported on VARTA element, pulse, pulse neo, link and flex storage devices

        registers = self._get_value_modbus(1066, 1)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.LITTLE
        ).decode_16bit_int()
        return result

    def get_apparent_power(self) -> int:
        # Apparent Power measured at the internal inverter. Positive = Charge,
        # Negative = Discharge
        # Supported on VARTA element, pulse, pulse neo, link and flex storage devices

        registers = self._get_value_modbus(1067, 1)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.LITTLE
        ).decode_16bit_int()
        return result

    def get_soc(self) -> int:
        # Current State of Charge of the Battery Power
        # Supported on VARTA element, pulse, pulse neo, link and flex storage devices

        registers = self._get_value_modbus(1068, 1)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.LITTLE
        ).decode_16bit_uint()
        return result

    def get_total_charged_energy(self) -> int:
        # Total charged energy
        # Supported on VARTA element, pulse, pulse neo, link and flex storage devices

        reg_low = self._get_value_modbus(1069, 1)
        reg_high = self._get_value_modbus(1070, 1)

        res_low = BinaryPayloadDecoder.fromRegisters(
            reg_low, Endian.BIG, Endian.LITTLE
        ).decode_16bit_uint()

        res_high = BinaryPayloadDecoder.fromRegisters(
            reg_high, Endian.BIG, Endian.LITTLE
        ).decode_16bit_uint()

        res = ((res_high << 16) | (res_low & 0xFFFF)) / 1000
        return res

    def get_installed_capacity(self) -> int:
        # Retrieves the total installed capacity in the device
        # Supported on VARTA element, pulse, pulse neo, link and flex storage devices

        registers = self._get_value_modbus(1071, 1)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.LITTLE
        ).decode_16bit_uint()
        # Installed capacity has to be multiplied by 10
        return result * 10

    def get_error_code(self) -> int:
        registers = self._get_value_modbus(1072, 1)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.LITTLE
        ).decode_16bit_uint()
        return result

    def get_grid_power(self) -> int:
        # Retrieves the current grid power measured at household grid connection point
        # Supported on VARTA element, pulse, pulse neo, link and flex storage devices

        registers = self._get_value_modbus(1078, 1)
        result = BinaryPayloadDecoder.fromRegisters(
            registers, Endian.BIG, Endian.LITTLE
        ).decode_16bit_int()
        return result

    def _get_value_modbus(self, address, count) -> list:
        if not self._modbus_client.is_socket_open():
            self._modbus_client.connect()

        try:
            rr = self._modbus_client.read_holding_registers(
                address, count, slave=self._slave
            )
        except ModbusException as exc:
            raise ValueError(ERROR_TEMPLATE.format(address)) from exc

        if rr.isError():
            raise ValueError(ERROR_TEMPLATE.format(address))

        return rr.registers

    @staticmethod
    def _clean_string(input_bytes: bytes) -> str:
        # I know this is super wierd. But i have no idea whats here going on in the
        # pymodbus library and i have to use this function to reformat the string
        # correctly
        return input_bytes[1::2].decode().replace("\x00", "")



import ast
import re
from dataclasses import dataclass, field
from typing import Any

from requests import Response, Session

ERROR_TEMPLATE = "An error occured while polling {}. Please check your connection"


@dataclass
class CgiData:
    info: dict = field(default_factory=dict)
    service: dict = field(default_factory=dict)
    ems: dict = field(default_factory=dict)
    energy: dict = field(default_factory=dict)


class CgiClient:
    def __init__(self, host, username=None, password=None):
        self.host = host
        self.username = username
        self.password = password

        self.session = Session()

    def get_all_data_cgi(self) -> CgiData:
        out = CgiData()
        out.energy = self.get_energy_cgi()
        out.service = self.get_service_cgi()
        out.ems = self.get_ems_cgi()
        out.info = self.get_info_cgi()
        return out

    def get_energy_cgi(self) -> dict[str, Any]:
        # get energy totals and charge load cycles from CGI
        # "EGrid_AC_DC": 0, "EGrid_DC_AC": 0, "EWr_AC_DC": 0, "EWr_DC_AC": 0,
        # "Chrg_LoadCycles": 0
        return self._get_cgi_as_dict("/cgi/energy.js")

    def get_ems_cgi(self) -> dict[str, Any]:
        # get ems data structure
        # usually a dict of 'wr': {...}, 'charger': [{...}], 'emeter': {...}, 'na': {}
        result: dict[str, Any] = {}

        conf = {
            key.lower(): value
            for key, value in self._get_cgi_as_dict("/cgi/ems_conf.js").items()
        }
        data = {
            key.lower(): value
            for key, value in self._get_cgi_as_dict("/cgi/ems_data.js").items()
        }

        for conf_key, conf_value in conf.items():
            data_key = conf_key.replace("conf", "data")
            if data_key not in data:
                continue

            data_value = data[data_key]
            if len(conf_value) == len(data_value):
                result[conf_key.replace("_conf", "")] = {
                    conf_value[i]: data_value[i] for i in range(0, len(conf_value))
                }
            elif len(data_value) >= 1 and isinstance(data_value[0], list):
                # exception for charger values, this is a list of values
                data_values: list[dict[str, Any]] = []
                for element in data_value:
                    if len(conf_value) != len(element):
                        continue
                    data_values.append(
                        {conf_value[i]: element[i] for i in range(0, len(conf_value))}
                    )
                result[conf_key.replace("_conf", "")] = data_values

        return result

    def get_service_cgi(self) -> dict[str, Any]:
        # get service and maintenance data from CGI
        # "FilterZeit": 0, "Fan": 0, "Main": 0
        return self._get_cgi_as_dict("/cgi/user_serv.js")

    def get_info_cgi(self) -> dict[str, Any]:
        # get various informations by the cgi/info.js
        return self._get_cgi_as_dict("/cgi/info.js")

    def _get_cgi_as_dict(self, path: str) -> dict[str, Any]:
        result = {}
        try:
            response = self._request_data(path)
            response.raise_for_status()
        except Exception as e:
            raise ValueError(ERROR_TEMPLATE.format(path)) from e

        tmp_list = response.text.replace("\n", "").split(";")
        value_list = [value for value in tmp_list if "=" in value]

        for value in value_list:
            splitted = value.split("=", 1)
            result[splitted[0].strip()] = ast.literal_eval(splitted[1].strip())

        return result

    def _request_data(self, urlEnding) -> Response:
        try:
            url = f"http://{self.host}{urlEnding}"
            # Check if a password is set
            if self.password:
                # Password is set so we check if already logged in
                self._check_logged_in()
            return self.session.get(url, timeout=3)
        except Exception as e:
            raise ValueError(ERROR_TEMPLATE) from e

    def _check_logged_in(self):
        pass_url = f"http://{self.host}/cgi/login"
        response = self.session.get(pass_url, timeout=3)
        response.raise_for_status()

        values = re.compile("userlevel = ([0-9]+)")
        results = values.findall(response.text)
        if "2" in results:
            # already logged in
            return True

        login_data = {"user": self.username, "password": self.password}
        response = self.session.post(pass_url, login_data, timeout=3)
        response.raise_for_status()
        return response.status_code == 200
